@@ -74,6 +74,9 @@ struct _CoglFramebuffer
   int                 green_bits;
   int                 alpha_bits;
 
+  gboolean            dither_enabled;
+  CoglColorMask       color_mask;
+
   /* We journal the textured rectangles we want to submit to OpenGL so
    * we have an oppertunity to batch them together into less draw
    * calls. */
