@@ -83,11 +83,9 @@ _cogl_winsys_context_init (CoglContext *context, GError **error)
 {
   context->winsys = &_cogl_winsys_stub_dummy_ptr;
 
-  if (!_cogl_gl_check_version (error))
+  if (!_cogl_context_update_features (context, error))
     return FALSE;
 
-  _cogl_gl_update_features (context);
-
   memset (context->winsys_features, 0, sizeof (context->winsys_features));
 
   return TRUE;
